/*
    Authors:
        Pavel Březina <pbrezina@redhat.com>

    Copyright (C) 2015 Red Hat

    SBUS: Interface introspection

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SSSD_DBUS_ERRORS_H_
#define SSSD_DBUS_ERRORS_H_

#define SBUS_ERROR_INTERNAL "org.freedesktop.sssd.Error.Internal"
#define SBUS_ERROR_NOT_FOUND "org.freedesktop.sssd.Error.NotFound"
#define SBUS_ERROR_UNKNOWN_DOMAIN "org.freedesktop.sssd.Error.UnknownDomain"

#define SBUS_ERROR_DP_FATAL "org.freedesktop.sssd.Error.DataProvider.Fatal"
#define SBUS_ERROR_DP_OFFLINE "org.freedesktop.sssd.Error.DataProvider.Offline"
#define SBUS_ERROR_DP_NOTSUP "org.freedesktop.sssd.Error.DataProvider.NotSupported"

#endif /* SSSD_DBUS_ERRORS_H_ */
